@@ -18,10 +18,10 @@ import numpy as np
 from jax._src.numpy import lax_numpy as jnp
 from jax._src.numpy.vectorize import vectorize
 from jax import ad_util
-from jax import api
+from jax._src import api
 from jax import lax
 from jax import ops
-from jax import dtypes
+from jax._src import dtypes
 from jax.interpreters import xla
 from jax.interpreters import ad
 from jax.interpreters import batching
@@ -33,6 +33,7 @@ from jax._src.lax.lax import (
 from jax._src.lax import lax as lax_internal
 from jax.lib import lapack
 
+from jax.lib import cuda_linalg
 from jax.lib import cusolver
 from jax.lib import rocsolver
 
@@ -111,6 +112,25 @@ def eigh(x, lower: bool = True, symmetrize_input: bool = True):
   v, w = eigh_p.bind(x, lower=lower)
   return v, w
 
+
+def lu_pivots_to_permutation(pivots, permutation_size: int):
+  """Converts the pivots (row swaps) returned by LU to a permutation.
+
+  We build a permutation rather than applying `pivots` directly to the rows
+  of a matrix because lax loops aren't differentiable.
+
+  Args:
+    pivots: an int32 array of shape (..., k) of row swaps to perform
+    permutation_size: the size of the output permutation. Has to be >= k.
+
+  Returns:
+    An int32 array of shape (..., permutation_size).
+  """
+  permutation = lu_pivots_to_permutation_p.bind(
+      pivots, permutation_size=int(permutation_size))
+  return permutation
+
+
 def lu(x):
   """LU decomposition with partial pivoting.
 
@@ -729,6 +749,100 @@ if rocsolver is not None:
   xla.backend_specific_translations['gpu'][triangular_solve_p] = \
       partial(_triangular_solve_gpu_translation_rule, rocsolver.trsm)
 
+# Support operation for LU decomposition: Transformation of the pivots returned
+# by LU decomposition into permutations.
+
+
+# Define this outside lu_pivots_to_permutation to ensure fori_loop cache hits
+def _lu_pivots_body_fn(i, permutation_and_swaps):
+  permutation, swaps = permutation_and_swaps
+  batch_dims = swaps.shape[:-1]
+  j = swaps[..., i]
+  iotas = jnp.ix_(*(lax.iota(jnp.int32, b) for b in batch_dims))
+  x = permutation[..., i]
+  y = permutation[iotas + (j,)]
+  permutation = ops.index_update(permutation, ops.index[..., i], y)
+  return ops.index_update(permutation, ops.index[iotas + (j,)], x), swaps
+
+
+@partial(api.jit, static_argnums=(1,))
+def _generic_lu_pivots_to_permutation(swaps, m):
+  """Converts the pivots (row swaps) returned by LU to a permutation.
+
+  We build a permutation rather than applying `swaps` directly to the rows
+  of a matrix because lax loops aren't differentiable.
+
+  Args:
+    swaps: an array of shape (..., k) of row swaps to perform
+    m: the size of the output permutation. m should be >= k.
+  Returns:
+    An int32 array of shape (..., m).
+  """
+  assert len(swaps.shape) >= 1
+  batch_dims = swaps.shape[:-1]
+  k = swaps.shape[-1]
+
+  permutation = lax.broadcasted_iota(jnp.int32, batch_dims + (m,),
+                                     len(batch_dims))
+  if m == 0:
+    return permutation
+  result, _ = lax.fori_loop(np.array(0, np.int32), np.array(k, np.int32),
+                            _lu_pivots_body_fn, (permutation, swaps))
+  return result
+
+
+def _lu_pivots_to_permutation_abstract_eval(pivots, *, permutation_size):
+  pivots = raise_to_shaped(pivots)
+  if isinstance(pivots, ShapedArray):
+    if pivots.ndim < 1 or pivots.dtype != np.dtype(np.int32):
+      raise ValueError(
+          'Argument to lu_pivots_to_permutation must have rank >= 1 and dtype '
+          'int32. Got shape={} and dtype={}'.format(pivots.shape, pivots.dtype))
+
+    if permutation_size < pivots.shape[-1]:
+      raise ValueError(
+          'Output permutation size {} has to exceed the trailing dimension of '
+          'the pivots. Got shape {}'.format(permutation_size, pivots.shape))
+
+    batch_dims = pivots.shape[:-1]
+    permutations = pivots.update(shape=batch_dims + (permutation_size,))
+  else:
+    permutations = pivots
+
+  return permutations
+
+
+def _lu_pivots_to_permutation_batching_rule(batched_args, batch_dims, *,
+                                            permutation_size):
+  x, = batched_args
+  bd, = batch_dims
+  x = batching.moveaxis(x, bd, 0)
+  return lu_pivots_to_permutation_p.bind(
+      x, permutation_size=permutation_size), 0
+
+
+def _lu_pivots_to_permutation_translation_rule(c, pivots, *, permutation_size):
+  lowered_fun = xla.lower_fun(
+      lambda x: _generic_lu_pivots_to_permutation(x, permutation_size),
+      multiple_results=False)
+  return lowered_fun(c, pivots)
+
+
+lu_pivots_to_permutation_p = Primitive('lu_pivots_to_permutation')
+lu_pivots_to_permutation_p.multiple_results = False
+lu_pivots_to_permutation_p.def_impl(
+    partial(xla.apply_primitive, lu_pivots_to_permutation_p))
+lu_pivots_to_permutation_p.def_abstract_eval(
+    _lu_pivots_to_permutation_abstract_eval)
+batching.primitive_batchers[lu_pivots_to_permutation_p] = (
+    _lu_pivots_to_permutation_batching_rule)
+xla.translations[lu_pivots_to_permutation_p] = (
+    _lu_pivots_to_permutation_translation_rule)
+
+if cuda_linalg:
+  xla.backend_specific_translations['gpu'][lu_pivots_to_permutation_p] = (
+      cuda_linalg.lu_pivots_to_permutation)
+
 # LU decomposition
 
 # Computes a pivoted LU decomposition such that
@@ -933,44 +1047,6 @@ if rocsolver is not None:
 xla.backend_specific_translations['tpu'][lu_p] = _lu_tpu_translation_rule
 
 
-# Define this outside lu_pivots_to_permutation to ensure fori_loop cache hits
-def _lu_pivots_body_fn(i, permutation_and_swaps):
-  permutation, swaps = permutation_and_swaps
-  batch_dims = swaps.shape[:-1]
-  j = swaps[..., i]
-  iotas = jnp.ix_(*(lax.iota(jnp.int32, b) for b in batch_dims))
-  x = permutation[..., i]
-  y = permutation[iotas + (j,)]
-  permutation = ops.index_update(permutation, ops.index[..., i], y)
-  return ops.index_update(permutation, ops.index[iotas + (j,)], x), swaps
-
-
-@partial(api.jit, static_argnums=(1,))
-def lu_pivots_to_permutation(swaps, m):
-  """Converts the pivots (row swaps) returned by LU to a permutation.
-
-  We build a permutation rather than applying `swaps` directly to the rows
-  of a matrix because lax loops aren't differentiable.
-
-  Args:
-    swaps: an array of shape (..., k) of row swaps to perform
-    m: the size of the output permutation. m should be >= k.
-  Returns:
-    An int32 array of shape (..., m).
-  """
-  assert len(swaps.shape) >= 1
-  batch_dims = swaps.shape[:-1]
-  k = swaps.shape[-1]
-
-  permutation = lax.broadcasted_iota(jnp.int32, batch_dims + (m,),
-                                     len(batch_dims))
-  if m == 0:
-    return permutation
-  result, _ = lax.fori_loop(np.array(0, np.int32), np.array(k, np.int32),
-                            _lu_pivots_body_fn, (permutation, swaps))
-  return result
-
-
 @partial(vectorize, excluded={3}, signature='(n,n),(n),(n,k)->(n,k)')
 def _lu_solve_core(lu, permutation, b, trans):
   m = lu.shape[0]
