@@ -22,9 +22,9 @@ import numpy as np
 from jax import lax
 from jax import core
 from jax import numpy as jnp
-from jax import dtypes
+from jax._src import dtypes
 from jax.core import NamedShape
-from jax.api import jit, vmap
+from jax._src.api import jit, vmap
 from jax._src.numpy.lax_numpy import _constant_like, _convert_and_clip_integer, asarray
 from jax.lib import xla_bridge
 from jax.lib import xla_client
@@ -63,8 +63,9 @@ def PRNGKey(seed: int) -> jnp.ndarray:
   # This breaks JIT invariance of PRNGKey for large ints, but supports the
   # common use-case of instantiating PRNGKey with Python hashes in X32 mode.
   if isinstance(seed, int):
-    seed = np.int64(seed)
-  seed_arr = jnp.asarray(seed)
+    seed_arr = jnp.asarray(np.int64(seed))
+  else:
+    seed_arr = jnp.asarray(seed)
   if seed_arr.shape:
     raise TypeError(f"PRNGKey seed must be a scalar; got {seed!r}.")
   if not np.issubdtype(seed_arr.dtype, np.integer):
@@ -120,7 +121,8 @@ def _threefry2x32_abstract_eval(*args):
                     .format(args))
   if all(isinstance(arg, core.ShapedArray) for arg in args):
     shape = lax._broadcasting_shape_rule(*args)
-    aval = core.ShapedArray(shape, jnp.dtype(jnp.uint32))
+    named_shape = core.join_named_shapes(*(a.named_shape for a in args))
+    aval = core.ShapedArray(shape, jnp.dtype(jnp.uint32), named_shape=named_shape)
   else:
     aval = core.UnshapedArray(jnp.dtype(jnp.uint32))
   return (aval,) * 2
@@ -562,7 +564,7 @@ def _shuffle(key, x, axis) -> jnp.ndarray:
   # another analysis (where the keys are generated one bit at a time).
   exponent = 3  # see tjablin@'s analysis for explanation of this parameter
   uint32max = jnp.iinfo(np.uint32).max
-  num_rounds = int(np.ceil(exponent * np.log(x.size) / np.log(uint32max)))
+  num_rounds = int(np.ceil(exponent * np.log(max(1, x.size)) / np.log(uint32max)))
 
   for _ in range(num_rounds):
     key, subkey = split(key)
@@ -1542,7 +1544,7 @@ def double_sided_maxwell(key: jnp.ndarray,
   return _double_sided_maxwell(key, loc, scale, shape, dtype)
 
 
-@partial(jit, static_argnums=(1, 2, 3, 4))
+@partial(jit, static_argnums=(3, 4))
 def _double_sided_maxwell(key, loc, scale, shape, dtype):
   params_shapes = lax.broadcast_shapes(np.shape(loc), np.shape(scale))
   if not shape:
@@ -1586,7 +1588,7 @@ def weibull_min(key: jnp.ndarray,
   return _weibull_min(key, scale, concentration, shape, dtype)
 
 
-@partial(jit, static_argnums=(1, 2, 3, 4))
+@partial(jit, static_argnums=(3, 4))
 def _weibull_min(key, scale, concentration, shape, dtype):
   random_uniform = uniform(
       key=key, shape=shape, minval=0, maxval=1, dtype=dtype)
