@@ -51,4 +51,4 @@ def logcdf(x, loc=0, scale=1):
 
 @_wraps(osp_stats.norm.ppf, update_doc=False)
 def ppf(q, loc=0, scale=1):
-  return jnp.array(special.ndtri(q) * scale + loc, 'float64')
+  return jnp.asarray(special.ndtri(q) * scale + loc, float)
