@@ -19,6 +19,7 @@ import operator
 import types
 from typing import Any, Callable
 
+from absl import logging
 import numpy as np
 
 from jax.config import config
@@ -182,7 +183,7 @@ def cache(max_size=4096):
       if config.jax_check_tracer_leaks:
         return f(*args, **kwargs)
       else:
-        return cached(bool(config.x64_enabled), *args, **kwargs)
+        return cached(config._trace_context(), *args, **kwargs)
 
     wrapper.cache_clear = cached.cache_clear
     wrapper.cache_info = cached.cache_info
@@ -196,7 +197,7 @@ def memoize(f):
 
   @functools.wraps(f)
   def wrapper(*args, **kwargs):
-    return memoized(bool(config.x64_enabled), *args, **kwargs)
+    return memoized(config._trace_context(), *args, **kwargs)
 
   wrapper.cache_clear = memoized.cache_clear
   wrapper.cache_info = memoized.cache_info
@@ -208,7 +209,7 @@ def prod(xs):
     out *= x
   return out
 
-class WrapHashably(object):
+class WrapHashably:
   __slots__ = ["val"]
 
   def __init__(self, val):
@@ -220,7 +221,7 @@ class WrapHashably(object):
   def __eq__(self, other):
     return self.val is other.val
 
-class Hashable(object):
+class Hashable:
   __slots__ = ["val"]
 
   def __init__(self, val):
@@ -232,6 +233,18 @@ class Hashable(object):
   def __eq__(self, other):
     return self.val == other.val
 
+class WrapKwArgs:
+  __slots__ = ["val"]
+
+  def __init__(self, val):
+    self.val = val
+
+  def __hash__(self):
+    return hash(tuple((k, v) for k, v in sorted(self.val.items())))
+
+  def __eq__(self, other):
+    return self.val == other.val
+
 def get_module_functions(module):
   """Finds functions in module.
   Args:
@@ -380,3 +393,22 @@ def maybe_named_axis(axis, if_pos, if_named):
   except TypeError:
     named = True
   return if_named(axis) if named else if_pos(pos)
+
+def distributed_debug_log(*pairs):
+  """Format and log `pairs` if config.jax_distributed_debug is enabled.
+
+  Args:
+    pairs: A sequence of label/value pairs to log. The first pair is treated as
+    a heading for subsequent pairs.
+  """
+  if config.jax_distributed_debug:
+    lines = ["\nDISTRIBUTED_DEBUG_BEGIN"]
+    try:
+      lines.append(f"{pairs[0][0]}: {pairs[0][1]}")
+      for label, value in pairs[1:]:
+        lines.append(f"  {label}: {value}")
+    except Exception as e:
+      lines.append("DISTRIBUTED_DEBUG logging failed!")
+      lines.append(f"{e}")
+    lines.append("DISTRIBUTED_DEBUG_END")
+    logging.warning("\n".join(lines))
