@@ -18,24 +18,26 @@ import string
 from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Union
 
 import jax
-from jax import ad_util, api, api_util, config
+from jax import ad_util, api_util, config
+from jax._src import api
 from jax import core, custom_derivatives, dtypes
 from jax import linear_util as lu
 from jax import numpy as jnp
 from jax import random, tree_util
 from jax._src import util
+from jax._src.lax import control_flow as lax_control_flow
+from jax._src.lax import fft as lax_fft
+from jax._src.lax import lax
+from jax._src.lax import linalg as lax_linalg
+import jax._src.random
 from jax.api_util import flatten_fun
-from jax.interpreters import ad, batching
-from jax.interpreters import masking
+from jax.interpreters import ad
 from jax.interpreters import pxla
 from jax.interpreters import sharded_jit
 from jax.interpreters import xla
-from jax._src.lax import lax
-from jax._src.lax import linalg as lax_linalg
-from jax._src.lax import control_flow as lax_control_flow
-from jax._src.lax import fft as lax_fft
-import jax._src.random
-from jax.lib import xla_bridge as xb
+from jax.lib import xla_client
+
+from . import shape_poly
 
 import numpy as np
 import tensorflow as tf  # type: ignore[import]
@@ -47,8 +49,8 @@ from tensorflow.compiler.xla import xla_data_pb2  # type: ignore[import]
 from tensorflow.compiler.xla.experimental.xla_sharding import xla_sharding  # type: ignore[import]
 # pylint: enable=g-direct-tensorflow-import
 
-from jax.lib import xla_client
 
+PolyShape = shape_poly.PolyShape
 
 # The scope name need to be a valid TensorFlow name. See
 # https://github.com/tensorflow/tensorflow/blob/r2.3/tensorflow/core/framework/node_def_util.cc#L731
@@ -81,8 +83,9 @@ def _is_tfval(v: TfVal) -> bool:
 def _safe_convert_to_tensor(val, dtype=None) -> TfVal:
   dtype = dtype if dtype else (val.dtype if hasattr(val, "dtype") else None)
   conversion_type = to_tf_dtype(dtype) if dtype else None
-  # We can convert directly, because all dtypes (even bfloat16) are the same
-  # in JAX and TF.
+  # The float0 type is not known to TF.
+  if dtype and dtype == dtypes.float0:
+    val = np.zeros(np.shape(val), conversion_type.as_numpy_dtype)
   return tf.convert_to_tensor(val, dtype=conversion_type)
 
 
@@ -119,7 +122,8 @@ def _xla_path_disabled_error(primitive_name: str) -> Exception:
 
 @functools.partial(api_util.api_hook, tag="jax2tf_convert")
 def convert(fun: Callable, *,
-            in_shapes: Optional[Sequence[Any]]=None,
+            polymorphic_shapes: Optional[Sequence[Any]]=None,
+            in_shapes=None,  # DEPRECATED
             with_gradient=True, enable_xla=True) -> Callable:
   """Transforms `fun` to be executed by TensorFlow.
 
@@ -128,10 +132,46 @@ def convert(fun: Callable, *,
 
   Args:
     fun: Function to be transformed. Its arguments and return value should be
-      JAX arrays, or (nested) standard Python containers (tuple/list/dict)
-      thereof.
+      JAX arrays, or nested standard Python containers (tuple/list/dict)
+      thereof (pytrees).
+
+    polymorphic_shapes: Specifies input shapes to be treated polymorphically
+      during conversion.
+
+      .. warning::
+      The shape-polymorphic conversion is an experimental feature. It is meant
+      to be sound, but it is known to reject some JAX programs that are
+      shape polymorphic. The details of this feature can change.
+
+      It should be a Python object with the same pytree structure as,
+      or a prefix of, the tuple of arguments to the function,
+      but with a shape specification corresponding to each argument.
+      The default value is `None`, which is a shortcut for a tuple of `None`
+      one for each argument, denoting that all shapes are monomorphic.
+      See [how optional parameters are matched to arguments](https://jax.readthedocs.io/en/latest/pytrees.html#applying-optional-parameters-to-pytrees).
+
+      A shape specification for an array argument
+      should be an object `PolyShape(dim0, dim1, ..., dimn)`
+      where each `dim` is a dimension specification: a positive integer denoting
+      a monomorphic dimension of the given size,
+      or a string denoting a dimension variable assumed to range over non-zero
+      dimension sizes,
+      or the special placeholder string "_" denoting a monomorphic dimension
+      whose size is given by the actual argument.
+      As a shortcut, an Ellipsis suffix in the
+      list of dimension specifications stands for a list of "_" placeholders.
+      For convenience, a shape specification can also be given as a string
+      representation, e.g.: "batch, ...", "batch, height, width, _", possibly
+      with surrounding parentheses: "(batch, ...)".
+
+      The conversion fails if it cannot ensure that the it would produce the same
+      sequence of TF ops for any non-zero values of the dimension variables.
+
+      See [the README](https://github.com/google/jax/blob/master/jax/experimental/jax2tf/README.md#shape-polymorphic-conversion)
+      for more details.
+
+    in_shapes: DEPRECATED in favor of `polymorphic_shapes`.
 
-    in_shapes: Not used. This is kept for backwards compatibility (see #6080).
     with_gradient: if set, will add a tf.custom_gradient to the converted
       function, by converting the ``jax.vjp(fun)``. Only first-order
       differentiation is supported for now. If the converted function is
@@ -173,19 +213,25 @@ def convert(fun: Callable, *,
     # tf.custom_gradient on functions that take a flat argument list.
     args_flat, in_tree = tree_util.tree_flatten((args, {}))
 
-    if in_shapes is None:
-      in_shapes_ = (None,) * len(args)
+    if polymorphic_shapes is None:
+      polymorphic_shapes_ = (None,) * len(args)
     else:
-      raise ValueError("The `in_shapes` parameter is not supported")
+      if not isinstance(polymorphic_shapes, Sequence) or len(args) != len(polymorphic_shapes):
+        msg = ("polymorphic_shapes must be a sequence with the same length as the argument list "
+               f"({len(args)}). Got polymorphic_shapes={polymorphic_shapes}.")
+        raise TypeError(msg)
+      polymorphic_shapes_ = tuple(polymorphic_shapes)
 
     # Expand the in_shapes to match the argument pytree
-    in_shapes_flat = tuple(api_util.flatten_axes("jax2tf.convert in_shapes",
-                                                 in_tree.children()[0], in_shapes_))
+    polymorphic_shapes_flat = tuple(api_util.flatten_axes("jax2tf.convert polymorphic_shapes",
+                                                          in_tree.children()[0],
+                                                          polymorphic_shapes_))
 
     # Construct the abstract values for the flat arguments, possibly based on
     # the input shapes and the in_shapes if given. May create new shape
     # variables.
-    args_avals_flat = _input_avals(args_flat, in_shapes_flat)
+    args_avals_flat, shapeenv = _args_to_avals_and_env(args_flat,
+                                                       polymorphic_shapes_flat)
 
     f = lu.wrap_init(fun)
     # out_tree_thunk() will be the output tree, after running _interpret_fun.
@@ -209,25 +255,25 @@ def convert(fun: Callable, *,
         _, pullback_jax = jax.vjp(fun, *args_jax)
         return pullback_jax(out_cts_jax)
 
-      if in_shapes is None:
-        vjp_in_shapes = None
+      if polymorphic_shapes is None:
+        vjp_polymorphic_shapes = None
       else:
-        args_in_shapes = tree_util.tree_unflatten(in_tree.children()[0], in_shapes_flat)
-        out_cts_in_shapes = tree_util.tree_unflatten(
+        args_polymorphic_shapes = tree_util.tree_unflatten(in_tree.children()[0], polymorphic_shapes_flat)
+        out_cts_polymorphic_shapes = tree_util.tree_unflatten(
           out_tree_thunk(),
           tuple(str(out_aval.shape) for out_aval in _out_cts_avals))  # type: ignore
-        vjp_in_shapes = [args_in_shapes, out_cts_in_shapes]
+        vjp_polymorphic_shapes = [args_polymorphic_shapes, out_cts_polymorphic_shapes]
       out_cts = tree_util.tree_unflatten(out_tree_thunk(), out_cts_flat)
       # TODO: enable higher-order gradients
       with tf.name_scope("jax2tf_vjp"):
         in_cts = convert(fun_vjp_jax, with_gradient=False,
-                         in_shapes=vjp_in_shapes)(args, out_cts)
+                         polymorphic_shapes=vjp_polymorphic_shapes)(args, out_cts)
       return in_cts
 
     try:
       global _shape_env
       assert not _shape_env, f"Unexpected shape environment {_shape_env}"
-      _shape_env = _make_shape_env(args_avals_flat, args_flat)
+      _shape_env = shapeenv
 
       if with_gradient:
         @tf.custom_gradient
@@ -254,6 +300,7 @@ def convert(fun: Callable, *,
 
   return converted_fun
 
+
 # Internals
 
 
@@ -321,19 +368,10 @@ def _interpret_jaxpr(jaxpr: core.ClosedJaxpr, *args: TfVal) -> Sequence[TfVal]:
 
 ### tracer
 
-PolyDim = Union[int, masking.Poly]  # A polymorphic shape dimension
-
-def _poly_dim_to_tf_dim(dim: PolyDim) -> Optional[int]:
-  if isinstance(dim, int):
-    return dim
-  elif dim.is_constant:
-    return int(dim)
-  else:
-    return None
-
 def _aval_to_tf_shape(aval: core.AbstractValue) -> Tuple[Optional[int], ...]:
   """Generate a TF shape, possibly containing None for polymorphic dimensions."""
-  return tuple(map(_poly_dim_to_tf_dim, aval.shape))  # type: ignore[attr-defined]
+  return tuple(map(lambda d: None if isinstance(d, shape_poly.DimVar) else d,
+                   aval.shape))  # type: ignore[attr-defined]
 
 
 def _tfval_shape_dtype(val: TfVal) -> Tuple[Sequence[Optional[int]], DType]:
@@ -351,129 +389,54 @@ def _tfval_shape_dtype(val: TfVal) -> Tuple[Sequence[Optional[int]], DType]:
     return raw_aval.shape, raw_aval.dtype  # type: ignore[attr-defined]
 
 
-def _input_avals(args: Sequence[TfVal], in_shapes: Sequence[Optional[str]]) -> Sequence[core.AbstractValue]:
-  """Abstract values for the input arguments."""
+# A dimension environment maps dimension variables to TF expressions that
+# compute the value of the dimension. These expressions refer to the TF
+# function arguments.
+_ShapeEnv = Dict[shape_poly.DimVar, TfVal]
+def _args_to_avals_and_env(args: Sequence[TfVal],
+                           polymorphic_shapes: Sequence[Optional[Union[str, PolyShape]]]) -> \
+  Tuple[Sequence[core.AbstractValue], _ShapeEnv]:
+  """Computes abstract values and a dimension environment for arguments.
 
-  def input_aval(arg: TfVal, in_shape: Optional[str]) -> core.AbstractValue:
+  Args:
+    args: the arguments, TF inputs.
+    polymorphic_shapes: the polymorphic specifications for the arguments.
+
+  Returns: a tuple of a sequence of abtract values corresponding to the arguments
+    and a dimension environment.
+  """
+  shapeenv: _ShapeEnv = {}
+  def input_aval(arg: TfVal, polymorphic_shape: Optional[str]) -> core.AbstractValue:
     """The abstract value for an input."""
     raw_shape, dtype = _tfval_shape_dtype(arg)
 
-    if in_shape is None:
-      if any(d is None for d in raw_shape):
-        msg = ("in_shape must be specified when the argument "
-               f"shape {raw_shape} is partially known.")
-        raise TypeError(msg)
+    aval_shape = shape_poly.parse_spec(polymorphic_shape, raw_shape)
+
+    for i, d in enumerate(aval_shape):
+      if type(d) is int:
+        assert d == np.shape(arg)[i]
+      elif type(d) is shape_poly.DimVar and d not in shapeenv:
+        # Even if the shape of `arg` is known, we still use `tf.shape` for
+        # safety, because the promise is that we will convert the function
+        # to work for any value of the dimension.
+        shapeenv[d] = tf.shape(arg)[i]  # type: ignore[index]
       else:
-        return core.ShapedArray(raw_shape, dtype)
+        # TODO: add an assertion tf.shape(arg)[i] == env[d]
+        pass
 
-    in_shape_spec = masking.parse_spec(in_shape)
-    if len(in_shape_spec) != len(raw_shape):
-      msg = (f"in_shape {in_shape} has different rank than actual argument "
-             f"shape {raw_shape}")
-      raise TypeError(msg)
-    try:
-      # TODO: improve finalize_spec error reporting, so we don't have to code our own
-      shape = masking.finalize_spec(in_shape_spec, raw_shape)
-    except TypeError:
-      msg = (f"in_shape {in_shape} has `_` placeholders for argument shape "
-             f"dimensions that are unknown: {raw_shape}")
-      raise TypeError(msg)
+    return core.ShapedArray(aval_shape, dtype)
 
-    for dim_idx, (s, raw_s) in enumerate(util.safe_zip(shape, raw_shape)):
-      s_int: Optional[int] = _poly_dim_to_tf_dim(s)
-      if s_int != raw_s and s_int is not None:
-        msg = (f"in_shape {in_shape} (resolved to {shape}) does not match "
-               f"argument shape {raw_shape} in dimension {dim_idx}")
-        raise TypeError(msg)
-
-    return core.ShapedArray(shape, dtype)
-
-  return tuple(map(input_aval, args, in_shapes))
+  avals = tuple(map(input_aval, args, polymorphic_shapes))  # type: ignore
+  return avals, shapeenv
 
 # A shape environment maps shape variables to TfVal.
-ShapeEnv = Dict[str, TfVal]
-_shape_env = {}  # type: ShapeEnv
+_shape_env = {}  # type: _ShapeEnv
 
-def _eval_shape(shape: Sequence[PolyDim]) -> Sequence[TfVal]:
+def _eval_shape(shape: Sequence[shape_poly.DimSize]) -> Sequence[TfVal]:
   assert all(map(lambda x: x is not None, shape)), (
       f"Argument shape should be a valid JAX shape but got {shape}")
-  return masking.eval_poly_shape(shape, _shape_env)
-
-# Extracting a shape environment by solving the shape variables.
-# The shape environment will be derived by using `tf.shape` from the
-# dynamic shape of arguments, not their static shape.
-def _make_shape_env(args_avals: Sequence[core.AbstractValue],
-                    args: Sequence[TfVal]) -> Dict[str, TfVal]:
-  eqns = [(p, tf.shape(a)[i])
-          for a_aval, a in util.safe_zip(args_avals, args)
-          for i, p in enumerate(a_aval.shape)]
-  return _solve_shape_vars(eqns)
-
-ShapeEqn = Tuple[PolyDim, TfVal]
-def _solve_shape_vars(eqns: Sequence[ShapeEqn]) -> Dict[str, TfVal]:
-  """Solves a number of equations "poly = tfval" into an shape environment."""
-  # A simple variable elimination algorithm for now
-  solved: Dict[str, TfVal] = {}  # Already solved vars
-
-  def simplify_poly(p: PolyDim) -> Optional[Union[TfVal,
-                                                  Tuple[str, TfVal, TfVal]]]:
-    # Simplifies polynomial given `solved`
-    # Returns (v, f, rest) such that p = v * f + rest, or
-    #         rest such that p = rest, or None
-    v = None
-    f = None
-    rest = []
-    if isinstance(p, int):
-      return tf.constant(p)
-    for m, m_factor in p.items():
-      simpl_m: Union[str, TfVal] = simplify_mon(m, p)
-      if isinstance(simpl_m, str):  # A var
-        if v is not None:
-          return None
-        v = simpl_m
-        f = m_factor
-      else:  # A value
-        rest.append(tf.math.multiply(simpl_m, m_factor))
-    rest_val = functools.reduce(tf.math.add, rest, tf.constant(0))
-    return rest_val if v is None else (v, f, rest_val)
-
-  def simplify_mon(m: masking.Mon, in_poly: masking.Poly) -> Union[str, TfVal]:
-    # Simplifies monomial given `solved`
-    # Returns either a variable, or a solved value
-    if not m:
-      return tf.constant(1)
-    if m.degree > 1:
-      msg = ("only linear polynomials are supported as input shape "
-             f"specifications. Found '{m}' in '{in_poly}'.")
-      raise TypeError(msg)
-    var = list(m.keys())[0]
-    return solved.get(var, var)
-
-  remaining = eqns
-  while remaining:
-    new_remaining = []
-    for eqn in remaining:
-      eq_p, eq_val = eqn
-      p_simpl = simplify_poly(eq_p)
-      if p_simpl is None:
-        new_remaining.append(eqn)
-        continue
-      if not isinstance(p_simpl, tuple):
-        # TODO: add an assertion rest == eq_v
-        continue
-      var, factor, rest = p_simpl
-      # p = var * factor + rest
-      # TODO: add an assertion eq_v >= rest and (eq_v - rest) mod factor == 0
-      solved[var] = tf.math.floordiv(tf.math.subtract(eq_val, rest), factor)
-
-    if len(new_remaining) < len(remaining):
-      remaining = new_remaining
-    else:
-      msg = "Cannot solve"
-      raise ValueError(msg)
-
-  return solved
-
+  return tuple(_shape_env[d] if type(d) is shape_poly.DimVar else d  # type: ignore[index]
+               for d in shape)
 
 def shape_as_value(x):
   """Injects the shape of `x` as an array value.
@@ -485,78 +448,80 @@ def shape_as_value(x):
 
      jnp.sum(x) / np.prod(jax2tf.shape_as_value(x))
   """
-  return shape_as_value_p.bind(x)
+  # return shape_as_value_p.bind(x)
+  return NotImplementedError("shape_as_value is deprecated")
 
-# TODO: move this to masking or to some common library, if approved
-shape_as_value_p = core.Primitive("shape_as_value")
-shape_as_value_p.multiple_results = True
-def _shape_as_value_impl(x):
-  x_shape = np.shape(x)
-  def dim_to_int(dim: PolyDim) -> int:
-    dim_int = _poly_dim_to_tf_dim(dim)
-    if dim_int is None:
-      msg = ("shape_as_value is not implemented for non-constant shapes "
-             "except for masking and jax2tf. "
-             f"Has shape: {x_shape}")
-      raise TypeError(msg)
-    else:
-      return dim_int
-  return tuple(map(dim_to_int, x_shape))
 
-shape_as_value_p.def_impl(_shape_as_value_impl)
-
-def _shape_as_value_abstract(x_aval: core.AbstractValue) -> Sequence[core.AbstractValue]:
-  rank = len(x_aval.shape)  # type: ignore[attr-defined]
-  return (core.ShapedArray((), dtypes.canonicalize_dtype(np.int_), weak_type=True),) * rank
-
-shape_as_value_p.def_abstract_eval(_shape_as_value_abstract)
-
-def _shape_as_value_translation(comp, x):
-  return xla_client._xla.ops.Tuple(comp,
-                                   tuple(xb.constant(comp, d)
-                                         for d in comp.GetShape(x).dimensions()))
-
-xla.translations[shape_as_value_p] = _shape_as_value_translation
-
-def _shape_as_value_jvp_rule(primals, tangents):
-  # The shape does not depend on the contents of the input
-  x, = primals
-  zero = ad.Zero.from_value(0.)
-  return shape_as_value(x), (zero,) * len(x.shape)
-
-ad.primitive_jvps[shape_as_value_p] = _shape_as_value_jvp_rule
-
-def _shape_as_value__batching_rule(batched_args, batch_dims):
-  xv, = batched_args
-  batch_dim, = batch_dims
-  batch_size = xv.shape[batch_dim]
-  batched_shape = shape_as_value(xv)
-  one_shape = batched_shape[0:batch_dim] + batched_shape[batch_dim+1:]
-  res = tuple(jnp.broadcast_to(d, (batch_size, 1)) for d in one_shape)
-  return res, (0,) * len(one_shape)
-
-batching.primitive_batchers[shape_as_value_p] = _shape_as_value__batching_rule
-
-def _shape_as_value_masking_rule(operands, operands_logical_shapes):
-  x_logical_shape, = operands_logical_shapes
-  return tuple(x_logical_shape)
-
-masking.masking_rules[shape_as_value_p] = _shape_as_value_masking_rule
-
-def _shape_as_value_tf(x: TfVal,
-                       _in_avals: Sequence[core.AbstractValue],
-                       _out_aval: core.AbstractValue) -> TfVal:
-  x_aval = _in_avals[0]
-  def dim_to_tfval(dim: PolyDim, dim_idx: int) -> TfVal:
-    dim_int = _poly_dim_to_tf_dim(dim)
-    if dim_int is not None:
-      return tf.convert_to_tensor(dim_int)
-    else:
-      return tf.shape(x)[dim_idx]
-  return tuple(dim_to_tfval(dim, dim_idx)
-               for dim_idx, dim in enumerate(x_aval.shape))  # type: ignore[attr-defined]
-
-tf_impl_with_avals[shape_as_value_p] = _shape_as_value_tf
+# # TODO: move this to masking or to some common library, if approved
+# shape_as_value_p = core.Primitive("shape_as_value")
+# shape_as_value_p.multiple_results = True
+# def _shape_as_value_impl(x):
+#   x_shape = np.shape(x)
+#   def dim_to_int(dim: shape_poly.DimSize) -> int:
+#     dim_int = _poly_dim_to_tf_dim(dim)
+#     if dim_int is None:
+#       msg = ("shape_as_value is not implemented for non-constant shapes "
+#              "except for masking and jax2tf. "
+#              f"Has shape: {x_shape}")
+#       raise TypeError(msg)
+#     else:
+#       return dim_int
+#   return tuple(map(dim_to_int, x_shape))
+#
+# shape_as_value_p.def_impl(_shape_as_value_impl)
+#
+# def _shape_as_value_abstract(x_aval: core.AbstractValue) -> Sequence[core.AbstractValue]:
+#   rank = len(x_aval.shape)  # type: ignore[attr-defined]
+#   return (core.ShapedArray((), dtypes.canonicalize_dtype(np.int_), weak_type=True),) * rank
+#
+# shape_as_value_p.def_abstract_eval(_shape_as_value_abstract)
+#
+# def _shape_as_value_translation(comp, x):
+#   return xla_client._xla.ops.Tuple(comp,
+#                                    tuple(xb.constant(comp, d)
+#                                          for d in comp.GetShape(x).dimensions()))
+#
+# xla.translations[shape_as_value_p] = _shape_as_value_translation
+#
+# def _shape_as_value_jvp_rule(primals, tangents):
+#   # The shape does not depend on the contents of the input
+#   x, = primals
+#   zero = ad.Zero.from_value(0.)
+#   return shape_as_value(x), (zero,) * len(x.shape)
+#
+# ad.primitive_jvps[shape_as_value_p] = _shape_as_value_jvp_rule
+#
+# def _shape_as_value__batching_rule(batched_args, batch_dims):
+#   xv, = batched_args
+#   batch_dim, = batch_dims
+#   batch_size = xv.shape[batch_dim]
+#   batched_shape = shape_as_value(xv)
+#   one_shape = batched_shape[0:batch_dim] + batched_shape[batch_dim+1:]
+#   res = tuple(jnp.broadcast_to(d, (batch_size, 1)) for d in one_shape)
+#   return res, (0,) * len(one_shape)
+#
+# batching.primitive_batchers[shape_as_value_p] = _shape_as_value__batching_rule
+#
+# def _shape_as_value_masking_rule(operands, operands_logical_shapes):
+#   x_logical_shape, = operands_logical_shapes
+#   return tuple(x_logical_shape)
+#
+# masking.masking_rules[shape_as_value_p] = _shape_as_value_masking_rule
+#
+# def _shape_as_value_tf(x: TfVal,
+#                        _in_avals: Sequence[core.AbstractValue],
+#                        _out_aval: core.AbstractValue) -> TfVal:
+#   x_aval = _in_avals[0]
+#   def dim_to_tfval(dim: shape_poly.DimSize, dim_idx: int) -> TfVal:
+#     dim_int = _poly_dim_to_tf_dim(dim)
+#     if dim_int is not None:
+#       return tf.convert_to_tensor(dim_int)
+#     else:
+#       return tf.shape(x)[dim_idx]
+#   return tuple(dim_to_tfval(dim, dim_idx)
+#                for dim_idx, dim in enumerate(x_aval.shape))  # type: ignore[attr-defined]
+#
+# tf_impl_with_avals[shape_as_value_p] = _shape_as_value_tf
 
 # TODO(b/26854495): pylint doesn't understand slots and inheritance.
 # pylint: disable=assigning-non-slot
@@ -602,16 +567,15 @@ class TensorFlowTracer(core.Tracer):
         assert aval_dtype == val_dtype, f"expected {aval_dtype} == {val_dtype}"
         for aval_dim, val_dim in util.safe_zip(self._aval.shape, val_shape):  # type: ignore[attr-defined]
           if val_dim is None:
-            assert isinstance(aval_dim, masking.Poly), f"expected {self._aval.shape} == {val_shape}"  # type: ignore[attr-defined]
-          elif not isinstance(aval_dim, masking.Poly):
+            assert isinstance(aval_dim, shape_poly.DimVar), f"expected {self._aval.shape} == {val_shape}"  # type: ignore[attr-defined]
+          elif not isinstance(aval_dim, shape_poly.DimVar):
             assert aval_dim == val_dim, f"expected {self._aval.shape} == {val_shape}"  # type: ignore[attr-defined]
           else:
-            # We have a TF value with known shape, and the abstract shape is a polynomial
-            # As an extra check, verify the value if the shape env are only constants
+            # We have a TF value with known shape, and the abstract shape is a shape variable.
             try:
-              aval_int = int(masking.eval_poly(aval_dim, _shape_env))
+             aval_int = int(_eval_shape([aval_dim]))  # type: ignore
             except TypeError:
-              continue
+             continue
             assert aval_int == val_dim, f"expected {self._aval.shape} == {val_shape}. Found {aval_int} != {val_dim}."  # type: ignore
 
       self.val = val
@@ -774,9 +738,8 @@ class TensorFlowTrace(core.Trace):
 
 def to_tf_dtype(jax_dtype):
   if jax_dtype == dtypes.float0:
-    return tf.float32
-  else:
-    return tf.dtypes.as_dtype(jax_dtype)
+    jax_dtype = dtypes.bfloat16
+  return tf.dtypes.as_dtype(jax_dtype)
 
 def to_jax_dtype(tf_dtype):
   return tf_dtype.as_numpy_dtype
@@ -794,12 +757,14 @@ tf_not_yet_impl = [
 
   "igamma_grad_a",
   "random_gamma_grad",
+  "reduce_precision",
 
   # Not high priority?
   "after_all", "all_to_all", "create_token",
   "infeed", "outfeed", "pmax_p",
   "pmin", "ppermute", "psum", "pmax", "pgather",
   "axis_index", "pdot", "all_gather",
+  "lu_pivots_to_permutation",
   "rng_bit_generator",
 
   "xla_pmap",
@@ -1083,12 +1048,13 @@ def _bitcast_convert_type(operand, new_dtype):
 tf_impl[lax.bitcast_convert_type_p] = _bitcast_convert_type
 
 
-def _clamp(minval, operand, maxval):
+def _clamp(minval, operand, maxval, *, _in_avals, _out_aval):
   # The below permits mirroring the behavior of JAX when maxval < minval
-  maxval = tf.broadcast_to(maxval, operand.shape)
-  minval = tf.math.minimum(tf.broadcast_to(minval, operand.shape), maxval)
+  op_shape_tf_val = _eval_shape(_in_avals[1].shape)
+  maxval = tf.broadcast_to(maxval, op_shape_tf_val)
+  minval = tf.math.minimum(tf.broadcast_to(minval, op_shape_tf_val), maxval)
   return tf.clip_by_value(operand, minval, maxval)
-tf_impl[lax.clamp_p] = _clamp
+tf_impl_with_avals[lax.clamp_p] = _clamp
 
 
 def _concatenate(*operands, dimension):
@@ -1230,7 +1196,7 @@ def _try_tf_conv(lhs, rhs, window_strides, padding, lhs_dilation, rhs_dilation,
 def _conv_general_dilated(lhs, rhs, window_strides, padding, lhs_dilation,
                           rhs_dilation, dimension_numbers, feature_group_count,
                           batch_group_count, lhs_shape, rhs_shape, precision,
-                          _in_avals, _out_aval):
+                          preferred_element_type, _in_avals, _out_aval):
   """Implementation of lax.conv_general_dilated_p using XlaConv."""
   if not _enable_xla:
     info_or_result = _try_tf_conv(
@@ -1262,7 +1228,7 @@ def _dot_general(lhs, rhs, dimension_numbers, precision, preferred_element_type)
   del precision
   del preferred_element_type
   (lhs_contracting, rhs_contracting), (lhs_batch, rhs_batch) = dimension_numbers
-  lhs_dim, rhs_dim = len(lhs.shape), len(rhs.shape)
+  lhs_ndim, rhs_ndim = len(lhs.shape), len(rhs.shape)
   # This condition ensures that:
   # 1) the considered dtype is not tf.bfloat16/tf.int32, which are supported by
   #    tf.linalg.einsum but not by tf.linalg.matmul;
@@ -1275,9 +1241,9 @@ def _dot_general(lhs, rhs, dimension_numbers, precision, preferred_element_type)
   #    matrix/matrix, vector/matrix or matrix/vector multiplication.
   if (not lhs.dtype in [tf.bfloat16, tf.int32]
       and lhs_batch == rhs_batch == tuple(range(len(lhs_batch)))
-      and lhs_dim - rhs_dim in [-1, 0, 1]
-      and 1 <= lhs_dim - len(lhs_batch) <= 2
-      and 1 <= rhs_dim - len(rhs_batch) <= 2
+      and lhs_ndim - rhs_ndim in [-1, 0, 1]
+      and 1 <= lhs_ndim - len(lhs_batch) <= 2
+      and 1 <= rhs_ndim - len(rhs_batch) <= 2
       and lhs_contracting == (len(lhs.shape) - 1,)
       and rhs_contracting == (len(lhs_batch),)):
     # All the inputs to tf.linalg.matmul must have 2 inner dimensions,
@@ -1299,14 +1265,15 @@ def _dot_general(lhs, rhs, dimension_numbers, precision, preferred_element_type)
     #     and the resulting shape is (). We need to squeeze the result of
     #     tf.linalg.matmul as it will have shape [1, 1].
     squeeze_idxs = []
-    if lhs_dim - len(lhs_batch) == 1:
-      lhs = tf.expand_dims(lhs, lhs_dim - 1)
+    if lhs_ndim - len(lhs_batch) == 1:
+      lhs = tf.expand_dims(lhs, lhs_ndim - 1)
       squeeze_idxs.append(len(lhs.shape) - 2)
-    if rhs_dim - len(rhs_batch) == 1:
-      rhs = tf.expand_dims(rhs, rhs_dim - 2)
+    if rhs_ndim - len(rhs_batch) == 1:
+      rhs = tf.expand_dims(rhs, rhs_ndim)
       squeeze_idxs.append(len(rhs.shape) - 1)
     result = tf.linalg.matmul(lhs, rhs)
     if len(squeeze_idxs) != 0:
+      assert all([result.shape[i] == 1 for i in squeeze_idxs])
       result = tf.squeeze(result, squeeze_idxs)
     return result
 
@@ -1386,6 +1353,8 @@ def _pad(operand, padding_value, *, padding_config,
   if not _enable_xla:
     raise _xla_path_disabled_error("pad")
   out = tfxla.pad(operand, padding_value, low, high, interior)
+  # TODO(b/184499027): improve shape inference for XlaPad
+  out.set_shape(_aval_to_tf_shape(_out_aval))
   return out
 tf_impl_with_avals[lax.pad_p] = _pad
 
@@ -1785,15 +1754,25 @@ def _gather(operand, start_indices, *, dimension_numbers, slice_sizes,
   return out
 tf_impl_with_avals[lax.gather_p] = _gather
 
-def _slice(operand, start_indices, limit_indices, strides):
+def _slice(operand, start_indices, limit_indices, strides,
+           _in_avals, _out_aval):
   if strides is None:
     strides = [1] * len(start_indices)
-  slices = tuple(map(slice, start_indices, limit_indices, strides))
-  return operand[slices]
-tf_impl[lax.slice_p] = _slice
+  slices = tuple(map(slice,
+                     _eval_shape(start_indices),
+                     _eval_shape(limit_indices),
+                     _eval_shape(strides)))
+  out = operand[slices]
+  # TODO(b/184503314): improve shape inference for __getitem__
+  out.set_shape(_aval_to_tf_shape(_out_aval))
+  return out
+
+tf_impl_with_avals[lax.slice_p] = _slice
 
 
-def _dynamic_slice(operand, *start_indices, slice_sizes):
+def _dynamic_slice(operand, *start_indices, slice_sizes,
+                   _in_avals: Sequence[core.ShapedArray],
+                   _out_aval: core.ShapedArray):
   # Here we could use tf.slice. Similarly, for lax.gather we can sometimes use
   # tf.gather. But those have different semantics for index-out-of-bounds than
   # JAX (and XLA). We have tried to force compilation, by wrapping into
@@ -1805,12 +1784,12 @@ def _dynamic_slice(operand, *start_indices, slice_sizes):
   if not _enable_xla:
     raise _xla_path_disabled_error("dynamic_slice")
   res = tfxla.dynamic_slice(operand, tf.stack(start_indices),
-                            size_indices=slice_sizes)
+                            size_indices=_eval_shape(slice_sizes))
   # TODO: implement shape inference for XlaDynamicSlice
-  res.set_shape(tuple(slice_sizes))
+  res.set_shape(_aval_to_tf_shape(_out_aval))
   return res
 
-tf_impl[lax.dynamic_slice_p] = _dynamic_slice
+tf_impl_with_avals[lax.dynamic_slice_p] = _dynamic_slice
 
 def _scatter_dimensions_proto(indices_shape, dimension_numbers):
   proto = xla_data_pb2.ScatterDimensionNumbers()
@@ -1972,7 +1951,6 @@ def _sort(*operands: TfVal, dimension: int, is_stable: bool,
   if not _enable_xla:
     raise _xla_path_disabled_error("sort")
   assert 1 <= num_keys <= len(operands)
-  assert all([operands[0].shape == op.shape for op in operands[1:]])
   assert 0 <= dimension < len(
       operands[0].shape
   ), f"Invalid {dimension} for ndim {len(operands[0].shape)}"
@@ -2081,20 +2059,20 @@ def _eig(operand: TfVal, compute_left_eigenvectors: bool,
 
 tf_impl[lax_linalg.eig_p] = _eig
 
-def _eigh(operand: TfVal, lower: bool):
+def _eigh(operand: TfVal, lower: bool, _in_avals, _out_aval):
   if operand.shape[-1] == 0:
-    v, w = operand, tf.reshape(operand, operand.shape[:-1])
+    v, w = operand, tf.reshape(operand, _eval_shape(_in_avals[0].shape[:-1]))
   else:
     if not lower:
       operand = tf.linalg.adjoint(operand)
     w, v = tf.linalg.eigh(operand)
-  cast_type = { tf.complex64: tf.float32
-              , tf.complex128: tf.float64 }.get(operand.dtype)
+  cast_type = { tf.complex64: tf.float32,
+                tf.complex128: tf.float64 }.get(operand.dtype)
   if cast_type is not None:
     w = tf.cast(w, cast_type)
   return v, w
 
-tf_impl[lax_linalg.eigh_p] = _eigh
+tf_impl_with_avals[lax_linalg.eigh_p] = _eigh
 
 def _lu(operand: TfVal, _in_avals, _out_aval):
   return _convert_jax_impl(lax_linalg._lu_python)(operand, _in_avals=_in_avals,
@@ -2104,9 +2082,13 @@ tf_impl_with_avals[lax_linalg.lu_p] = _lu
 
 def _triangular_solve(a: TfVal, b: TfVal, *, left_side: bool, lower: bool,
                       transpose_a: bool, conjugate_a: bool,
-                      unit_diagonal: bool):
+                      unit_diagonal: bool,
+                      _in_avals: Sequence[core.ShapedArray],
+                      _out_aval: core.ShapedArray):
   if unit_diagonal:
-    a = tf.linalg.set_diag(a, tf.ones(a.shape[:-1], dtype=a.dtype))
+    a_aval, _ = _in_avals
+    a_shape = _eval_shape(a_aval.shape)
+    a = tf.linalg.set_diag(a, tf.ones(a_shape[:-1], dtype=a.dtype))
   if not left_side:
     rank = len(a.shape)
     transpose_dimensions = list(range(rank - 2)) + [rank - 1, rank - 2]
@@ -2123,7 +2105,7 @@ def _triangular_solve(a: TfVal, b: TfVal, *, left_side: bool, lower: bool,
     result = tf.transpose(result, transpose_dimensions)
   return result
 
-tf_impl[lax_linalg.triangular_solve_p] = _triangular_solve
+tf_impl_with_avals[lax_linalg.triangular_solve_p] = _triangular_solve
 
 def _linear_solve(*args: TfVal, const_lengths, jaxprs, _in_avals, _out_aval):
   return _convert_jax_impl(lax_control_flow._custom_linear_solve_impl)(
