@@ -69,7 +69,7 @@ from jax.experimental import jax2tf
 from jax.interpreters import xla
 
 import numpy as np
-import tensorflow as tf
+import tensorflow as tf  # type: ignore[import]
 
 config.parse_flags_with_absl()
 
@@ -99,7 +99,7 @@ class JaxPrimitiveTest(tf_test_util.JaxToTfTestCase):
   # If you want to run this test for only one harness, add parameter
   # `one_containing="foo"` to parameterized below.
   @primitive_harness.parameterized(
-      primitive_harness.all_harnesses, include_jax_unimpl=False,
+      primitive_harness.all_harnesses, include_jax_unimpl=False
       )
   @jtu.ignore_warning(
       category=UserWarning, message="Using reduced precision for gradient.*")
