@@ -27,6 +27,7 @@ from jax import tree_util
 from jax.config import config
 from jax.experimental import jax2tf
 from jax.interpreters import masking
+from jax._src import util
 import numpy as np
 import tensorflow as tf  # type: ignore[import]
 
@@ -44,6 +45,8 @@ def _make_tf_args(args):
 def _make_tf_input_signature(*tf_args) -> List[tf.TensorSpec]:
   # tf_args can be PyTrees
   def _make_one_arg_signature(tf_arg):
+    if np.isscalar(tf_arg):
+      tf_arg = np.array(tf_arg)
     return tf.TensorSpec(np.shape(tf_arg), tf_arg.dtype)
 
   return tf.nest.map_structure(_make_one_arg_signature, list(tf_args))
@@ -237,7 +240,7 @@ class JaxToTfTestCase(jtu.JaxTestCase):
 
   def CheckShapePolymorphism(self, f_jax: Callable, *,
                              input_signature: Sequence[tf.TensorSpec],
-                             in_shapes: Optional[Sequence[Any]],
+                             polymorphic_shapes: Optional[Sequence[Any]],
                              expected_output_signature: tf.TensorSpec):
     """Convert a function using polymorphic shapes.
 
@@ -248,30 +251,36 @@ class JaxToTfTestCase(jtu.JaxTestCase):
         must match the `input_signature`. (see jax2tf.convert).
     """
     f_tf = tf.function(
-        jax2tf.convert(f_jax, in_shapes=in_shapes),
+        jax2tf.convert(f_jax, polymorphic_shapes=polymorphic_shapes),
         autograph=False,
         input_signature=input_signature)
     concrete_f_tf = f_tf.get_concrete_function(*input_signature)
     if expected_output_signature:
+      # Strangely, output_shapes can be a single shape for a function with a
+      # single result, or a list/tuple of shapes.
       concrete_output_tf_shape = concrete_f_tf.output_shapes
-      assert not isinstance(concrete_output_tf_shape, tuple)  # A single result
-      self.assertEqual(
-          tuple(expected_output_signature.shape),
-          tuple(concrete_output_tf_shape))
+      if not isinstance(concrete_output_tf_shape, (tuple, list)):  # Single result
+        assert not isinstance(expected_output_signature, (tuple, list))
+        expected_output_signature = [expected_output_signature]
+        concrete_output_tf_shape = [concrete_output_tf_shape]
+
+      for expected, found in util.safe_zip(expected_output_signature,
+                                           concrete_output_tf_shape):
+        self.assertEqual(tuple(expected.shape), tuple(found))
     return f_tf
 
-  def MakeInputSignature(self, *in_shapes):
+  def MakeInputSignature(self, *polymorphic_shapes):
     """From a pytree of in_shape string specification, make a pytree of tf.TensorSpec.
 
     Dimension variables are replaced with None.
     """
 
-    def in_shape_to_tensorspec(in_shape: str) -> tf.TensorSpec:
-      in_spec = masking.parse_spec(in_shape)
+    def polymorphic_shape_to_tensorspec(poly_shape: str) -> tf.TensorSpec:
+      in_spec = masking.parse_spec(poly_shape)
       return tf.TensorSpec(
           tuple(
               int(dim_spec) if dim_spec.is_constant else None
               for dim_spec in in_spec),
           dtype=tf.float32)
 
-    return tree_util.tree_multimap(in_shape_to_tensorspec, in_shapes)
+    return tree_util.tree_multimap(polymorphic_shape_to_tensorspec, polymorphic_shapes)
