@@ -15,7 +15,7 @@
 """Common functions for neural network libraries."""
 
 # flake8: noqa: F401
-from .scale import (
+from jax._src.image.scale import (
   resize,
   ResizeMethod,
   scale_and_translate,
