@@ -20,7 +20,7 @@ from typing import Any, Callable, Dict
 from . import partial_eval as pe
 from ..config import config
 from .. import core
-from ..dtypes import dtype, float0
+from .._src.dtypes import dtype, float0
 from ..core import (Trace, Tracer, get_aval, call_p, Primitive, Literal,
                     raise_to_shaped)
 from ..ad_util import (add_jaxvals, add_jaxvals_p, zeros_like_jaxval, zeros_like_aval,
@@ -176,8 +176,8 @@ def backward_pass(jaxpr: core.Jaxpr, consts, primals_in, cotangents_in):
     ct_env[v] = add_tangents(ct_env[v], ct) if v in ct_env else ct
     if config.jax_enable_checks:
       ct_aval = core.get_aval(ct_env[v])
-      joined_aval = core.lattice_join(v.aval, ct_aval).strip_weak_type()
-      assert v.aval.strip_weak_type() == joined_aval, (prim, v.aval, ct_aval)
+      joined_aval = core.lattice_join(v.aval, ct_aval).strip_weak_type().strip_named_shape()
+      assert v.aval.strip_weak_type().strip_named_shape() == joined_aval, (prim, v.aval, ct_aval)
 
   def read_cotangent(v):
     return ct_env.pop(v, Zero(v.aval))
@@ -285,11 +285,11 @@ class JVPTrace(Trace):
     if 'name' in params:
       params = dict(params, name=wrap_name(params['name'], 'jvp'))
     f_jvp = jvp_subtrace(f, self.main)
+    f_jvp, nz_tangents_out = nonzero_tangent_outputs(f_jvp)
     if isinstance(call_primitive, core.MapPrimitive):
       in_axes = params['in_axes']
       tangent_in_axes = [ax for ax, nz in zip(in_axes, nz_tangents) if nz]
       out_axes_thunk = params['out_axes_thunk']
-      f_jvp, nz_tangents_out = nonzero_tangent_outputs(f_jvp)
       # The new thunk depends deterministically on the old thunk and the wrapped function.
       # Any caching already has to include the wrapped function as part of the key, so we
       # only use the previous thunk for equality checks.
@@ -304,7 +304,8 @@ class JVPTrace(Trace):
                     out_axes_thunk=new_out_axes_thunk)
     f_jvp, out_tree_def = traceable(f_jvp, len(primals), tangent_tree_def)
     update_params = call_param_updaters.get(call_primitive)
-    new_params = update_params(params, nz_tangents) if update_params else params
+    new_params = (update_params(params, nz_tangents, nz_tangents_out)
+                  if update_params else params)
     result = call_primitive.bind(f_jvp, *primals, *nonzero_tangents, **new_params)
     primal_out, tangent_out = tree_unflatten(out_tree_def(), result)
     return [JVPTracer(self, p, t) for p, t in zip(primal_out, tangent_out)]
@@ -471,13 +472,12 @@ def add_tangents(x, y):
     return add_jaxvals(x, y)
 
 
-def defbilinear_broadcasting(bcast, prim, lhs_rule, rhs_rule):
+def defbilinear(prim, lhs_rule, rhs_rule):
   assert isinstance(prim, Primitive)
-  lhs_jvp = lambda g, x, y, **kwargs: prim.bind(bcast(g, y), y, **kwargs)
-  rhs_jvp = lambda g, x, y, **kwargs: prim.bind(x, bcast(g, x), **kwargs)
+  lhs_jvp = lambda g, x, y, **kwargs: prim.bind(g, y, **kwargs)
+  rhs_jvp = lambda g, x, y, **kwargs: prim.bind(x, g, **kwargs)
   defjvp(prim, lhs_jvp, rhs_jvp)
   primitive_transposes[prim] = partial(bilinear_transpose, lhs_rule, rhs_rule)
-defbilinear: Callable = partial(defbilinear_broadcasting, lambda g, x: g)
 
 def bilinear_transpose(lhs_rule, rhs_rule, cotangent, x, y, **kwargs):
   assert is_undefined_primal(x) ^ is_undefined_primal(y)
