@@ -23,7 +23,7 @@ from weakref import ref
 import numpy as np
 
 from .. import core
-from .. import dtypes
+from .._src import dtypes
 from .. import linear_util as lu
 from ..ad_util import Zero
 from .._src.util import (unzip2, safe_zip, safe_map, toposort, partial,
@@ -400,8 +400,8 @@ call_partial_eval_rules: Dict[core.Primitive, Callable] = {}
 call_param_updaters: Dict[core.Primitive, Callable] = {}
 
 
-def abstract_eval_fun(fun, *avals, **params):
-  _, avals_out, _ = trace_to_jaxpr_dynamic(lu.wrap_init(fun, params), avals)
+def abstract_eval_fun(fun, *avals, transform_name="", **params):
+  _, avals_out, _ = trace_to_jaxpr_dynamic(lu.wrap_init(fun, params), avals, transform_name)
   assert all(isinstance(aval, AbstractValue) for aval in avals_out)
   return avals_out
 
