@@ -15,8 +15,11 @@
 # This module is largely a wrapper around `jaxlib` that performs version
 # checking on import.
 
+import os
+from typing import Optional
+
 __all__ = [
-  'cuda_prng', 'cusolver', 'rocsolver', 'jaxlib', 'lapack',
+  'cuda_linalg', 'cuda_prng', 'cusolver', 'rocsolver', 'jaxlib', 'lapack',
   'pocketfft', 'pytree', 'tpu_client', 'version', 'xla_client'
 ]
 
@@ -69,6 +72,11 @@ try:
 except ImportError:
   cusolver = None
 
+try:
+  from jaxlib import cusparse  # pytype: disable=import-error
+except ImportError:
+  cusparse = None
+
 try:
   from jaxlib import rocsolver  # pytype: disable=import-error
 except ImportError:
@@ -79,6 +87,11 @@ try:
 except ImportError:
   cuda_prng = None
 
+try:
+  from jaxlib import cuda_linalg  # pytype: disable=import-error
+except ImportError:
+  cuda_linalg = None
+
 # Jaxlib code is split between the Jax and the Tensorflow repositories.
 # Only for the internal usage of the JAX developers, we expose a version
 # number that can be used to perform changes without breaking the master
@@ -89,3 +102,9 @@ try:
   from jaxlib import tpu_client  # pytype: disable=import-error
 except:
   tpu_client = None
+
+
+cuda_path: Optional[str]
+cuda_path = os.path.join(os.path.dirname(jaxlib.__file__), "cuda")
+if not os.path.isdir(cuda_path):
+  cuda_path = None
