@@ -28,11 +28,11 @@ from absl.testing import parameterized
 import numpy as np
 import numpy.random as npr
 
-from . import api
+from ._src import api
 from . import core
-from . import dtypes as _dtypes
+from ._src import dtypes as _dtypes
 from . import lax
-from .config import flags, bool_env, config
+from ._src.config import flags, bool_env, config
 from ._src.util import partial, prod
 from .tree_util import tree_multimap, tree_all, tree_map, tree_reduce
 from .lib import xla_bridge
@@ -52,6 +52,14 @@ flags.DEFINE_integer(
   int(os.getenv('JAX_NUM_GENERATED_CASES', 10)),
   help='Number of generated cases to test')
 
+flags.DEFINE_integer(
+  'max_cases_sampling_retries',
+  int(os.getenv('JAX_MAX_CASES_SAMPLING_RETRIES', 100)),
+  'Number of times a failed test sample should be retried. '
+  'When an unseen case cannot be generated in this many trials, the '
+  'sampling process is terminated.'
+)
+
 flags.DEFINE_bool(
     'jax_skip_slow_tests',
     bool_env('JAX_SKIP_SLOW_TESTS', False),
@@ -658,10 +666,13 @@ def rand_some_nan(rng):
       return base_rand(shape, dtype)
 
     dims = _dims_of_shape(shape)
-    nan_flips = rng.rand(*dims) < 0.1
+    r = rng.rand(*dims)
+    nan_flips = r < 0.1
+    neg_nan_flips = r < 0.05
 
     vals = base_rand(shape, dtype)
     vals = np.where(nan_flips, np.array(np.nan, dtype=dtype), vals)
+    vals = np.where(neg_nan_flips, np.array(-np.nan, dtype=dtype), vals)
 
     return _cast_to_shape(np.asarray(vals, dtype=dtype), shape, dtype)
 
@@ -794,6 +805,29 @@ def cases_from_gens(*gens):
     for i in range(cases_per_size):
       yield ('_{}_{}'.format(size, i),) + tuple(gen(size) for gen in gens)
 
+def named_cases_from_sampler(gen):
+  seen = set()
+  retries = 0
+  rng = npr.RandomState(42)
+  def choose_one(x):
+    if not isinstance(x, (list, tuple)):
+      x = list(x)
+    return [x[rng.randint(len(x))]]
+  while (len(seen) < FLAGS.num_generated_cases and
+         retries < FLAGS.max_cases_sampling_retries):
+    retries += 1
+    cases = list(gen(choose_one))
+    if not cases:
+      continue
+    if len(cases) > 1:
+      raise RuntimeError("Generator is expected to only return a single case when sampling")
+    case = cases[0]
+    if case["testcase_name"] in seen:
+      continue
+    retries = 0
+    seen.add(case["testcase_name"])
+    yield case
+
 
 class JaxTestLoader(absltest.TestLoader):
   def getTestCaseNames(self, testCaseClass):
